from typing import List, Union

from discord import Emoji

from discordmenu.embed.base import Box


class Text(Box):
    def __init__(self, value: Union[str, Box]):
        super().__init__(self)
        self.value = value

    def to_markdown(self) -> str:
        if isinstance(self.value, Box):
            return self.value.to_markdown()
        return self.value


class LabeledText(Box):
    def __init__(self, name: str, value: Union[Box, str]):
        super().__init__(self)
        self._name = BoldText(name)
        if isinstance(value, str):
            self._value = Text(value)
        else:
            self._value = value

    def to_markdown(self) -> str:
        return "{} {}".format(self._name.to_markdown(), self._value.to_markdown())

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = BoldText(name)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = Text(value)


class LinkedText(Box):
    def __init__(self, name: str, link: str):
        super().__init__(self)
        self._name = Text(name)
        self.link = link

    def to_markdown(self) -> str:
        return "[{}]({})".format(self._name.to_markdown(), self.link)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = Text(value)


class BoldText(Box):
    def __init__(self, value: Union[str, Text, LinkedText]):
        super().__init__(self)
        if isinstance(value, Text):
            self._value = Text(value.to_markdown())
        elif isinstance(value, LinkedText):
            self._value = Text(value.to_markdown())
        self._value = Text(value)

    def to_markdown(self) -> str:
        return "**{}**".format(self._value.to_markdown())

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = Text(value)


class HighlightableLinks(Box):
    def __init__(self, links: List[LinkedText], highlighted: int, delimiter=", "):
        super().__init__(self)
        self.delimiter = delimiter
        self.links = links
        self._highlighted = links[highlighted]

    def _get_link_markdown(self, link: LinkedText) -> str:
        return link.to_markdown() if link != self.highlighted else BoldText(link.name.value).to_markdown()

    def to_markdown(self) -> str:
        return self.delimiter.join([self._get_link_markdown(link) for link in self.links])

    @property
    def highlighted(self):
        return self._highlighted

    @highlighted.setter
    def highlighted(self, highlighted):
        if len(self.links) <= highlighted < 0:
            raise Exception("Selected is out of bounds")
        self.highlighted = self.links[highlighted]


class CustomEmoji(Box):
    def __init__(self, emoji: Emoji):
        super().__init__(self)
        self.emoji = emoji

    def to_markdown(self) -> str:
        return "<{}:{}:{}>".format("a" if self.emoji.animated else "", self.emoji.name, self.emoji.id)



from discordmenu.discord_client import send_embed_control
from discordmenu.embed.menu import EmbedMenu, EmbedControl
from discordmenu.emoji.emoji_cache import EmojiCache
from discordmenu.reaction_filter import ValidEmojiReactionFilter, NotPosterEmojiReactionFilter, \
    MessageOwnerReactionFilter
from discordmenu.test_embed import embed_view

c = EmojiCache([713430119343063182])


def embed_menu_1(original_author_id):
    panes = {
        '🧪': embed_control_1,
        'bcookie': embed_control_2,
    }

    valid_emoji_names = [e.name for e in c.custom_emojis] + list(panes.keys())
    reaction_filters = [
        ValidEmojiReactionFilter(valid_emoji_names),
        NotPosterEmojiReactionFilter(),
        MessageOwnerReactionFilter(original_author_id)
    ]

    return EmbedMenu(reaction_filters, panes)


def embed_control_1(prev_embed_control, **state):
    emoji_button_names = ['bcookie', 'bnotdis', 'blobthinkingfast', 'googledog', 'tsubaki']
    return EmbedControl([embed_view(None, 1, **state)], [c.get_by_name(e) for e in emoji_button_names])


def embed_control_2(prev_embed_control, **state):
    emoji_button_names = ['🧪', 'bnotdis', 'bloblul', 'googlecat', 'googlecatface']
    return EmbedControl([embed_view(None, 2, **state)], [c.get_by_name(e) for e in emoji_button_names])


async def mt(self, ctx):
    original_author_id = ctx.message.author.id
    state = {'original_author_id': original_author_id}
    c.refresh_from_discord_bot(self.bot)
    menu = embed_menu_1(original_author_id)
    ec = menu.get_embed_control('🧪', state=state)
    await send_embed_control(ctx, ec)



class EmbedMenuEmojiConfig:
    def __init__(self, delete_message="❌"):
        self.delete_message = delete_message

    def to_list(self):
        return [self.delete_message]


DEFAULT_EMBED_MENU_EMOJI_CONFIG = EmbedMenuEmojiConfig()



from typing import List

from discord import Embed

from discordmenu.embed.components import EmbedMain, EmbedAuthor, EmbedFooter, EmbedThumbnail, EmbedBodyImage, \
    EmbedField

HIDDEN_CHAR = "\u200b"


def _get_field_name(field, first_chunk):
    return field.name.value if first_chunk else field.continuation_title or HIDDEN_CHAR


class EmbedView:
    def __init__(self,
                 embed_main: EmbedMain,
                 embed_author: EmbedAuthor = None,
                 embed_thumbnail: EmbedThumbnail = None,
                 embed_body_image: EmbedBodyImage = None,
                 embed_fields: List[EmbedField] = None,
                 embed_footer: EmbedFooter = None
                 ):
        self.embed_author = embed_author
        self.embed_body_image = embed_body_image
        self.embed_thumbnail = embed_thumbnail
        self.embed_main = embed_main
        self.embed_fields = [f for f in embed_fields if f] if embed_fields else []
        self.embed_footer = embed_footer

    def to_embed(self) -> Embed:
        embed = Embed(**self.embed_main)
        embed.set_thumbnail(**self.embed_thumbnail) if self.embed_thumbnail else None
        embed.set_image(**self.embed_body_image) if self.embed_body_image else None
        embed.set_author(**self.embed_author) if self.embed_author else None
        embed.set_footer(**self.embed_footer) if self.embed_footer else None

        for field in self._chunk_embed_fields(1024):
            embed.add_field(**field)

        return embed

    def _chunk_embed_fields(self, chunk_size):
        chunks = []
        for field in self.embed_fields:
            remaining = field.value.to_markdown()
            first_chunk = True
            while len(remaining) > 0:
                field_name = _get_field_name(field, first_chunk)
                if len(remaining) <= chunk_size:
                    chunks.append(EmbedField(field_name, remaining, field.inline))
                    break

                left = remaining[:chunk_size]
                right = remaining[chunk_size:]

                delimiter_index = left.rfind(field.chunk_delimiter)
                if delimiter_index == -1:
                    raise Exception("Could not chunk by delimiter")

                left_body = left[:delimiter_index]

                chunks.append(EmbedField(field_name, left_body, field.inline))
                if first_chunk:
                    first_chunk = False

                remaining = left[delimiter_index + 1:] + right
        return chunks

    @staticmethod
    def from_message(existing_embed: Embed) -> "EmbedView":
        main = EmbedMain(existing_embed.title, existing_embed.url, existing_embed.colour, existing_embed.description)

        e_author = existing_embed.author
        author = EmbedAuthor(e_author.name, e_author.url, e_author.icon_url)

        thumbnail = EmbedThumbnail(existing_embed.thumbnail.url)
        body_image = EmbedBodyImage(existing_embed.image.url)

        e_footer = existing_embed.footer
        footer = EmbedFooter(e_footer.text, e_footer.icon_url)

        fields = [EmbedField(e.name, e.value, e.inline) for e in existing_embed.fields]
        return EmbedView(main, author, thumbnail, body_image, fields, footer)



from typing import List, Union

from discord import Color

from discordmenu.embed.base import CustomMapping, Box
from discordmenu.embed.text import Text


class EmbedField(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["name", "value", "inline"]

    def __init__(self, title: str, body: Union[Box, str], inline: bool = False, chunk_delimiter='\n',
                 continuation_title=''):
        self.continuation_title = continuation_title
        self.chunk_delimiter = chunk_delimiter
        self._name = Text(title)
        self._value = body
        self.inline = inline

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, title):
        self._name = Text(title)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, body: Box):
        self._value = Box(body)


class EmbedMain(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["url", "title", "color", "description"]

    def __init__(self, title: str = "", url: str = "", color: Union[str, Color] = Color.default(),
                 description: Union[Box, str] = ""):
        self.title = title
        self.url = url
        if isinstance(color, Color):
            self.color = color
        else:
            self.color = int(color, 16 if not color.startswith("0x") else 0)
        self.description = description


class EmbedFooter(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["icon_url", "text"]

    def __init__(self, value: Union[Box, str], icon_url: str = ""):
        self.icon_url = icon_url
        self.text = value


class EmbedAuthor(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["name", "url", "icon_url"]

    def __init__(self, name: str, url: str = "", icon_url: str = ""):
        self.name = name
        self.icon_url = icon_url
        self.url = url


class EmbedThumbnail(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["url"]

    def __init__(self, url: str):
        self.url = url


class EmbedBodyImage(CustomMapping):
    @property
    def fields(self) -> List[str]:
        return ["url"]

    def __init__(self, url: str):
        self.url = url



from collections.abc import Iterable
from typing import Optional

from discord import Reaction, Message, Member, RawReactionActionEvent

from discordmenu.embed.emoji import DEFAULT_EMBED_MENU_EMOJI_CONFIG, EmbedMenuEmojiConfig
from discordmenu.emoji.emoji import discord_emoji_to_emoji_name


class ReactionFilter:
    def __init__(self, reaction_filter: Optional["ReactionFilter"] = None):
        self.inner_filter: ReactionFilter = reaction_filter

    async def allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        parent = await self._allow_reaction(message, reaction, member)
        if parent:
            return True

        return await self.inner_filter.allow_reaction(message, reaction, member) if self.inner_filter else False

    async def allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        parent = await self._allow_reaction_raw(message, reaction)
        if parent:
            return True

        return await self.inner_filter.allow_reaction_raw(message, reaction) if self.inner_filter else False

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        return True

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        return True


class ValidEmojiReactionFilter(ReactionFilter):
    def __init__(self, valid_emoji_names: Iterable, default_emoji_override: EmbedMenuEmojiConfig = None,
                 filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        default_emojis = default_emoji_override or DEFAULT_EMBED_MENU_EMOJI_CONFIG
        emoji_set = set(default_emojis.to_list())
        emoji_set.update(valid_emoji_names)
        self.valid_emoji_names = list(emoji_set)

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        valid_emoji_reaction = discord_emoji_to_emoji_name(reaction.emoji) in self.valid_emoji_names
        return valid_emoji_reaction

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        valid_emoji_reaction = discord_emoji_to_emoji_name(reaction.emoji) in self.valid_emoji_names
        return valid_emoji_reaction


class NotPosterEmojiReactionFilter(ReactionFilter):
    """
    This prevents the bot from reacting to its own emojis in DM.
    """

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        return message.author.id != member.id

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        if not reaction.guild_id:
            return message.author.id != reaction.user_id
        return True


class MessageOwnerReactionFilter(ReactionFilter):
    def __init__(self, original_author_id: int, filters: Optional[ReactionFilter] = None):
        super().__init__(filters)
        self.original_author_id = original_author_id

    async def _allow_reaction(self, message: Message, reaction: Reaction, member: Member):
        return member.id == self.original_author_id

    async def _allow_reaction_raw(self, message: Message, reaction: RawReactionActionEvent):
        if reaction.guild_id:
            return reaction.member.id == self.original_author_id
        return True



from discord import Embed

from discordmenu.embed.base import Box
from discordmenu.embed.components import EmbedField, EmbedFooter
from discordmenu.embed.components import EmbedMain
from discordmenu.embed.components import EmbedThumbnail
from discordmenu.embed.menu import EmbedView
from discordmenu.intra_message_state import IntraMessageState
from discordmenu.embed.text import BoldText
from discordmenu.embed.text import HighlightableLinks
from discordmenu.embed.text import LabeledText
from discordmenu.embed.text import LinkedText
from discordmenu.embed.text import Text


def embed_view(embed_control, number, **state):
    view = EmbedView(
        EmbedMain(
            color="ffaaff",
            title="[3260] EMBED NUMBER {}".format(number),
            url="http://www.puzzledragonx.com/en/monster.asp?n=3260",
            description=Box(
                BoldText("Dragon/God")
            )
        ),
        embed_thumbnail=EmbedThumbnail("https://d1kpnpud0qoyxf.cloudfront.net/media/icons/03260.png"),
        embed_fields=[
            EmbedField(
                "Inheritable",
                Box(
                    LabeledText("Rarity", "7"),
                    LabeledText("Cost", "36"),
                ),
                inline=True
            ),
            EmbedField(
                "Stats",
                Box(
                    Text("HP 12037"),
                    Text("ATK 123"),
                ),
                inline=True
            ), EmbedField(
                "Active Skill",
                Text("Create 5 Fire and Light orbs; Charge all allies' skills by 1 turn"),
            ), EmbedField(
                "Alternate Evos",
                HighlightableLinks(
                    links=[
                        LinkedText("2141", "http://www.puzzledragonx.com/en/monster.asp?n=2141"),
                        LinkedText("2142", "http://www.puzzledragonx.com/en/monster.asp?n=2142"),
                        LinkedText("2970", "http://www.puzzledragonx.com/en/monster.asp?n=2970"),
                        LinkedText("3260", "http://www.puzzledragonx.com/en/monster.asp?n=3260"),
                    ],
                    highlighted=3
                )
            )
        ],
        embed_footer=EmbedFooter(
            "whawlekrjlkwer",
            IntraMessageState.serialize(
                "https://d1kpnpud0qoyxf.cloudfront.net/media/icons/05135.png",
                {"original_author_id": state['original_author_id']}
            )
        )
    )

    return view


def embed_1(reaction_add_event, message, **state):
    embed = Embed(
        **EmbedMain(
            color="ffaaff",
            title="[3260] Yamato Flame Dragon Caller, Tsubaki",
            url="http://www.puzzledragonx.com/en/monster.asp?n=3260",
            description=Box(
                BoldText("Dragon/God")
            )
        )
    )
    embed.set_thumbnail(url="https://d1kpnpud0qoyxf.cloudfront.net/media/icons/03260.png")
    e_field1 = EmbedField(
        "Inheritable",
        Box(
            LabeledText("Rarity", "7"),
            LabeledText("Cost", "36"),
        ),
        inline=True
    )
    e_field2 = EmbedField(
        "Stats",
        Box(
            Text("HP 12037"),
            Text("ATK 123"),
        ),
        inline=True
    )
    e_field = EmbedField(
        "Active Skill",
        Text("Create 5 Fire and Light orbs; Charge all allies' skills by 1 turn"),
    )
    e_field4 = EmbedField(
        "Alternate Evos",
        HighlightableLinks(
            links=[
                LinkedText("2141", "http://www.puzzledragonx.com/en/monster.asp?n=2141"),
                LinkedText("2142", "http://www.puzzledragonx.com/en/monster.asp?n=2142"),
                LinkedText("2970", "http://www.puzzledragonx.com/en/monster.asp?n=2970"),
                LinkedText("3260", "http://www.puzzledragonx.com/en/monster.asp?n=3260"),
            ],
            highlighted=3
        )
    )
    e_footer = EmbedFooter(
        "whawlekrjlkwer",
        IntraMessageState.serialize(
            "https://d1kpnpud0qoyxf.cloudfront.net/media/icons/05135.png",
            {"original_author_id": state['original_author_id']}
        )
    )
    embed.add_field(**e_field1)
    embed.add_field(**e_field2)
    embed.add_field(**e_field)
    embed.add_field(**e_field4)
    embed.set_footer(**e_footer)
    return embed


def embed_2(reaction_add_event, message, **state):
    embed = Embed(
        **EmbedMain(
            color="ffaaff",
            title="[3260] Yamato Flame Dragon Caller, THIS THING CHNGED",
            url="http://www.puzzledragonx.com/en/monster.asp?n=3260",
            description=Box(
                BoldText("Dragon/God"),
            )
        )
    )
    embed.set_thumbnail(url="https://d1kpnpud0qoyxf.cloudfront.net/media/icons/03260.png")
    e_field1 = EmbedField(
        "Inheritable",
        Box(
            LabeledText("Rarity", "7"),
            LabeledText("Cost", "36"),
        ),
        inline=True
    )
    e_field2 = EmbedField(
        "Stats",
        Box(
            Text("HP 12037"),
            Text("ATK 123"),
        ),
        inline=True
    )
    e_field = EmbedField(
        "Active Skill",
        Text("Create 5 Fire and Light orbs; Charge all allies' skills by 1 turn"),
    )
    e_field4 = EmbedField(
        "Alternate Evos",
        HighlightableLinks(
            links=[
                LinkedText("2141", "http://www.puzzledragonx.com/en/monster.asp?n=2141"),
                LinkedText("2142", "http://www.puzzledragonx.com/en/monster.asp?n=2142"),
                LinkedText("2970", "http://www.puzzledragonx.com/en/monster.asp?n=2970"),
                LinkedText("3260", "http://www.puzzledragonx.com/en/monster.asp?n=3260"),
            ],
            highlighted=3
        )
    )
    e_footer = EmbedFooter(
        "whawlekrjlkwer",
        IntraMessageState.serialize(
            "https://d1kpnpud0qoyxf.cloudfront.net/media/icons/05135.png",
            {"original_author_id": state['original_author_id']}
        )
    )

    embed.add_field(**e_field1)
    embed.add_field(**e_field2)
    embed.add_field(**e_field)
    embed.add_field(**e_field4)
    embed.set_footer(**e_footer)
    return embed



from typing import Callable, List, Optional, Coroutine, Dict

from discord import Message, RawReactionActionEvent

from discordmenu.discord_client import remove_reaction, update_embed_control, send_embed_control, \
    diff_emojis_raw
from discordmenu.embed.control import EmbedControl
from discordmenu.embed.emoji import DEFAULT_EMBED_MENU_EMOJI_CONFIG, EmbedMenuEmojiConfig
from discordmenu.embed.view_state import ViewState
from discordmenu.emoji.emoji import discord_emoji_to_emoji_name
from discordmenu.emoji.emoji_cache import emoji_cache
from discordmenu.reaction_filter import ReactionFilter

_IntraMessageState = Dict
NextEmbedControlFunc = Callable[[Optional[Message], Dict, _IntraMessageState], Coroutine[None, None, EmbedControl]]


class EmbedMenu:
    def __init__(self,
                 reaction_filters: List[ReactionFilter],
                 transitions: Dict[str, NextEmbedControlFunc],
                 initial_pane: Callable,
                 emoji_config: EmbedMenuEmojiConfig = DEFAULT_EMBED_MENU_EMOJI_CONFIG):
        self.emoji_config = emoji_config
        self.transitions = transitions
        self.reaction_filters = reaction_filters
        self.initial_pane = initial_pane

    async def create(self, ctx, state: ViewState):
        embed_control: EmbedControl = self.initial_pane(state)
        e_buttons = embed_control.emoji_buttons

        # Only add the close button if it doesn't exist, in case user has overridden it.
        e_buttons.append(
            self.emoji_config.delete_message) if self.emoji_config.delete_message not in e_buttons else None
        await send_embed_control(ctx, embed_control)

    async def transition(self, message, ims, emoji_clicked, **data):
        transition_func = self.transitions.get(emoji_clicked)
        if not transition_func:
            if emoji_clicked == discord_emoji_to_emoji_name(self.emoji_config.delete_message):
                await message.delete()
            return

        new_control = await transition_func(message, ims, **data)

        current_emojis = [e.emoji for e in message.reactions]
        next_emojis = new_control.emoji_buttons + [emoji_cache.get_emoji(self.emoji_config.delete_message)]

        emoji_diff = diff_emojis_raw(current_emojis, next_emojis)
        await update_embed_control(message, new_control, emoji_diff)

        if message.guild:
            await remove_reaction(message, emoji_clicked, ims['original_author_id'])

    async def should_respond_raw(self, message, event: RawReactionActionEvent):
        for reaction_filter in self.reaction_filters:
            allow = await reaction_filter.allow_reaction_raw(message, event)
            if not allow:
                return False
        return True

    async def should_respond(self, message, reaction, member):
        for reaction_filter in self.reaction_filters:
            allow = await reaction_filter.allow_reaction(message, reaction, member)
            if not allow:
                return False
        return True



from typing import List

from discord import Emoji as Emoji
from discord.ext.commands import Bot


class EmojiCache:
    def __init__(self, guild_ids: List[int]):
        self.guild_ids = guild_ids
        self.custom_emojis: List[Emoji] = []

    def set_guild_ids(self, guild_ids: List[int]):
        self.guild_ids = guild_ids

    def refresh_from_discord_bot(self, bot: Bot):
        self.custom_emojis = [e for g in bot.guilds if g.id in self.guild_ids for e in g.emojis]

    def refresh_from_emojis(self, emojis: List[Emoji]):
        self.custom_emojis = emojis

    def get_by_name(self, name):
        res = [e for e in self.custom_emojis if e.name == name]
        return res[0] if len(res) > 0 else name

    def get_emoji(self, name):
        for e in self.custom_emojis:
            if e.name == name:
                return str(e)
        return ":{}:".format(name)


emoji_cache = EmojiCache([])



from typing import List, Union

from discord import Message, Emoji

from discordmenu.embed.view import EmbedView


class EmbedControl:
    def __init__(self, embed_views: List[EmbedView], emoji_buttons: List[Union[Emoji, str]] = None):
        self.embed_views = embed_views
        self.emoji_buttons = emoji_buttons or []

    @staticmethod
    def from_message(message: Message) -> "EmbedControl":
        emojis = [r.emoji for r in message.reactions]
        embed_views = [EmbedView.from_message(e) for e in message.embeds]
        return EmbedControl(embed_views, emojis)



class ViewState:
    def __init__(self, original_author_id, menu_type, raw_query, extra_state=None):
        self.extra_state = extra_state or {}
        self.menu_type = menu_type
        self.original_author_id = original_author_id
        self.raw_query = raw_query

    def serialize(self):
        ret = {
            'raw_query': self.raw_query,
            'menu_type': self.menu_type,
            'original_author_id': self.original_author_id,
        }
        ret.update(self.extra_state)
        return ret



import base64
import json
from typing import Dict, Optional
from urllib.parse import urlparse, parse_qsl, urlencode, urlunparse, parse_qs

from discord import Embed

DEFAULT_QUERY_PARAM_KEYS = {
    'author': 'imsa',
    'image': 'imsi',
    'footer': 'imsf',
    'thumbnail': 'imst',
}


class IntraMessageState:
    @staticmethod
    def serialize(icon_url, json_dict, query_param_key='imsf') -> str:
        raw_bytes = base64.b64encode(json.dumps(json_dict).encode())
        data = str(raw_bytes)[2:-1]

        params = {query_param_key: data}
        url_parts = list(urlparse(icon_url))

        query = dict(parse_qsl(url_parts[4]))
        query.update(params)

        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)

    @staticmethod
    def deserialize(url, query_param_key) -> Optional[Dict]:
        parsed_url = urlparse(url)
        query_params_dict = parse_qs(parsed_url.query)
        data = query_params_dict.get(query_param_key)
        if not data:
            return None

        result_bytes = base64.b64decode(data[0])
        return json.loads(result_bytes)

    @staticmethod
    def extract_data(embed: Embed, query_param_keys: Dict = None) -> Dict:
        values_to_try = [
            ('author', embed.author.icon_url),
            ('image', embed.image.url),
            ('footer', embed.footer.icon_url),
            ('thumbnail', embed.thumbnail.url),
        ]

        if not query_param_keys:
            query_param_keys = DEFAULT_QUERY_PARAM_KEYS

        ret: Dict = {}
        for key, url in values_to_try:
            if not url:
                continue

            query_param_key = query_param_keys.get(key)
            data = IntraMessageState.deserialize(url, query_param_key)
            if data is None:
                # found url but not intra message state keys.
                continue

            IntraMessageState._merge_dicts(ret, data)
        return ret

    @staticmethod
    def _merge_dicts(target, data):
        for key, val in data.items():
            if key in target:
                existing_data = target[key]
                if not isinstance(existing_data, list):
                    # upgrade to vector
                    target[key] = [existing_data]
                target[key].append(val)
            else:
                target[key] = val



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="discord-menu",
    version="0.9.0",
    author="The Tsubotki Team",
    author_email="69992611+TsubakiBotPad@users.noreply.github.com",
    license="Apache-2.0 License",
    description="A library to create tabbed menus in Discord messages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TsubakiBotPad/discord-menu",
    packages=setuptools.find_packages(),
    install_requires=[""],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import asyncio
from typing import Dict, List, Optional

from discord import Embed, Emoji, Forbidden, Message, Reaction

from discordmenu.embed.control import EmbedControl
from discordmenu.emoji.emoji_cache import emoji_cache


async def update_message(message: Message, updated_messaged_contents, guild_message: bool,
                         emoji_diff: Dict[str, List[Emoji]] = None):
    if isinstance(updated_messaged_contents, Embed):
        await message.edit(embed=updated_messaged_contents)
    else:
        await message.edit(content=updated_messaged_contents)
    if emoji_diff:
        add = [message.add_reaction(e) for e in emoji_diff.get('add', [])]
        await asyncio.gather(*add)

        if guild_message:
            remove = [message.clear_reaction(e) for e in emoji_diff.get('remove', [])]
            await asyncio.gather(*remove)


async def remove_reaction(message: Message, emoji, user_id):
    if not message.guild:
        # bots don't have permission to delete reactions in DM. So check if theres a guild associated before attempting.
        return

    member = message.guild.get_member(user_id)

    try:
        await message.remove_reaction(emoji, member)
    except Forbidden:
        pass


async def send_embed_control(ctx, embed_control: "EmbedControl"):
    new_embed = embed_control.embed_views[0].to_embed()
    message = await ctx.send(embed=new_embed)

    emoji_to_add = [emoji_cache.get_by_name(e) for e in embed_control.emoji_buttons]
    add = [message.add_reaction(e) for e in emoji_to_add]
    await asyncio.gather(*add)


async def update_embed_control(message: Message, next_embed_control: Optional["EmbedControl"], emoji_diff: Dict):
    guild_message = bool(message.guild)

    if not next_embed_control:
        await message.delete()

    updated_message_contents = next_embed_control.embed_views[0].to_embed()
    await message.edit(embed=updated_message_contents)

    if emoji_diff:
        add = [message.add_reaction(e) for e in emoji_diff.get('add', [])]
        await asyncio.gather(*add)

        if guild_message:
            remove = [message.clear_reaction(e) for e in emoji_diff.get('remove', [])]
            await asyncio.gather(*remove)


def diff_emojis(message: Message, next_embed_control: Optional["EmbedControl"]):
    current_emojis = [e.emoji for e in message.reactions]
    next_emojis = next_embed_control.emoji_buttons
    return diff_emojis_raw(current_emojis, next_emojis)


def diff_emojis_raw(current_emojis: List[Reaction], next_emojis: List[Reaction]):
    return {
        'add': list(set(e for e in next_emojis if e and e not in current_emojis)),
        'remove': list(set(e for e in current_emojis if e and e not in next_emojis)),
    }
